#include <errno.h>
#include <stdio.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/types.h>

#define uint8_t unsigned char
#define int16_t short
#define uint16_t unsigned short

void send_and_recv(int fd2, uint8_t *buf, int slen, int rlen) {
	int r;
	if (slen != 0) {
	    r = write(fd2, buf, slen);
	    if (r != slen ) {
		fprintf(stderr,"cmd send error r = %d %02x\n", r , buf[0]);
		exit(1);
	    }
	}
	if (rlen != 0) {
	    r = read(fd2, buf , rlen);
	    if (r != rlen) {
		fprintf(stderr,"reply error r = %d %02x\n", r , buf[0]);
		exit(1);
	    }
	}
}

can_read(int fd) {
	int r;
	fd_set fds;
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = 0;
	FD_ZERO(&fds);
	FD_SET(fd, &fds);
	r = select(fd+1, &fds, NULL, NULL, &tv);
	return (r == 1);
}


init_usb910(int fd) {
	int r,i;
	uint8_t buf[1024];

printf("init_usb910\n");
	if (can_read(fd)) {
		printf("can read\n");
		r = read(fd, buf, 100);
		printf("init_usb910: read %d\n",r);
	}
	for (i=0; i<10; i++) {
		r = write(fd, "S\033", 2);
		if (r < 1) {
			perror("write 1");
			exit(1);
		}
		usleep(2000); /* 2 ms */
		if (!can_read(fd)) continue;
printf("read after cmd start\n");
		r = read(fd, buf, 100);
printf("read after cmd done\n");
		if (r == 7 && !strncmp(buf,"USB",3)) {
			buf[7] = 0;
			printf("nego OK [%s]\n", buf);
			r = write(fd, "a", 1);
			usleep(1000); /* 1 ms */
			r = read(fd, buf, 100);
			if (r == 1 && buf[0] == 'Y') {
				printf("auto address incliment OK\n");
				goto next;
			} else {
				printf("auto address incliment NG\n");
			}
		}
		printf("read len = %d\n",r);
	}
	return -1;
next:
	buf[0] = 'e';
	send_and_recv(fd, buf, 1, 1);
	if (buf[0] != 'Y') {
		fprintf(stderr,"enter extended mode Error\n",buf[0]);
	}
}

static int sseq;
static int seq_id;

test_send_req(int fd2, int count) {
	int r,i;
        uint8_t buf[1024];

	buf[0] = 's';	// send
	buf[1] = 9;	// echo
	buf[2] = count;	// send len
	for (i=0; i< count; i++) {
		buf[3+i] = seq_id;
	}

	r = write(fd2, buf, 3 + count);
	if (r != 3 + count) {
		fprintf(stderr,"write cmd error 1 r = %d \n", r);
		perror("write cmd 2");
		exit(1);
	}
}

test_send_repl(int fd2, int count) {
	int r,i;
        uint8_t buf[1024];
	r = read(fd2, buf , 1);
	if (r != 1 || buf[0] != count) {
		fprintf(stderr,"reply error 1 seq %d r = %d r[0] = %02x\n", sseq, r, buf[0]);
		exit(1);
	}
	sseq++;
}

test_recv_req(int fd2, int count) {
	int r,i,sum;
        uint8_t buf[1024];

	buf[0] = 'r';	// recv
	buf[1] = 9;	// test
	buf[2] = count;	// result len
	r = write(fd2, buf, 3);
	if (r != 3) {
		fprintf(stderr,"write cmd error 3 r = %d \n", r);
		perror("write cmd 3");
		exit(1);
	}
}

test_recv_repl(int fd2, int count, int chk) {
	int r,i,sum;
        uint8_t buf[1024];

	sum = 0;
retry:
	r = read(fd2, buf + sum, count + 1 - sum);
	if (r > 0) {
		sum += r;
		if (sum < count + 1) goto retry;
	}
	if ((r < 0) || (buf[0] != count)) {
		fprintf(stderr,"reply error 3 r = %d r[0] = %02x\n", r, buf[0]);
		exit(1);
	}
	for (i=0; i< chk && i <count; i++) {
		if (buf[1+i] != (seq_id & 0xff)) {
			fprintf(stderr,"check error (seq %d) %d \n", seq_id, buf[1+i]);
		}
	}
	seq_id++;
}

main(int argc,char *argv[]) {
	int fd,c,i,b1,b2,a1;
	int speed;
	struct termios term;
	uint8_t ibuf[1024];

	fd = open("/dev/ttyACM0", O_RDWR);
	if (fd < 0) {
		perror("open device");
		exit(1);
	}
	tcgetattr(fd, &term);
	//speed = B4800;
	speed = B19200;
	//speed = B57600;
	//speed = B115200;
	cfsetispeed(&term,speed);
	cfsetospeed(&term,speed);
	cfmakeraw(&term);
	usleep(100000); /* 100 ms */
	tcsetattr(fd, TCSANOW, &term);
	/* setup done */

	init_usb910(fd);
	a1 = 0;
	b1 = 4;
	b2 = 4;
	for (i=1; i< argc; i++) {
		if (argv[i][0] == 's') {
			b1 = atoi(argv[i]+1);
		} else if (argv[i][0] == 'r') {
			b2 = atoi(argv[i]+1);
		} else if (argv[i][0] == 'a') {
			a1 = atoi(argv[i]+1);
		}
	}
	printf("send %d bytes recv %d bytes x 1000 (pipelining %d) \n",b1,b2,a1);
	for (i=0; i<1000; i++) {
		if (b1) test_send_req(fd, b1);
		if (b1)	test_send_repl(fd, b1);
		if (b2) test_recv_req(fd, b2);
		if (b2)	test_recv_repl(fd, b2, b1);
	}
}
